import socket

STANDARD_WIDTH = 3.5

class Blinkenfoo(object):
    """This class represents a blinkenfoo instance.
    A blinkenfoo can display information send via upd and can listen to mqtt to play
    preconfigured animations. It als knows how many pixels it has:
    A pixel is one blinken-unit which can be controlled separately from the other
    units. A pixel can be a single LED on a i.E. ws2812 or a whole stripe.
    If the blinkenfoo is a led matrix then the pixel count is a*b.
    The physical width describes the width in cm of one blinkenunit/pixel. This is
    to make animations run the same speed on different blinkenfoo devices"""

    def __init__(self, p_name, p_udp_host, p_length, p_physikal_width = STANDARD_WIDTH, p_udp_port = 2390, p_mqtt_channel = None):
        self.name = p_name
        self.udp_host = p_udp_host
        self.udp_port = p_udp_port

        self.length = p_length
        self.physikal_width = p_physikal_width
        self.mqtt_channel = p_mqtt_channel
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._reset_buffer()

    def _unify_color_to_rgb(self, p_color):
        if type(p_color) is tuple:
            return p_color
        elif type(p_color) is str:
            return self._hex_to_rgb(p_color)
        else:
            raise ValueError("Unknown type! Color must be RGB Tuple or hex values as string")

    def _hex_string_to_rgb_list(self, p_hex_string):
        """
        Converts a hex string to an rgb frame
        """
        length = len(p_hex_string)
        if not length % 6 == 0:
            raise ValueError("Invalid Hex-string: length does not divide by 6")

        rgb_list = []
        for i in range(0, length, 6):
            hex_value = p_hex_string[i:i+6]
            rgb_list.append(self._hex_to_rgb(hex_value))

        return rgb_list

    def _hex_list_to_rgb_list(self, p_hex_list):
        """
        Converts a hex frame into a rgb frame
        """
        rgb_list = []

        for hex in p_hex_list:
            rgb_list.append(self._hex_to_rgb(hex))

        return rgb_list

    def _hex_to_rgb(self, p_hex):
        """Turns a hex value #123456 into rgb (18, 52, 86)"""
        return tuple(int(p_hex[i:i+2], 16) for i in (0, 2 ,4))

    """
    Example:
    package = _generate_packet(0, 9, [(255, 0, 0), (255, 0, 0), ...., (255, 0, 0)])
    _send_packet(package)
    Works for python2
    """
    def _generate_packet(self, startpixel, endpixel, rgb):
            """ Generates a udp packet the blinkenfoo understands"""
            pack = chr(0) + chr(startpixel) + chr(0) + chr(endpixel)
    #        pack = chr((startpixel >> 8) & 0xff) + chr(startpixel & 0xff) + chr((startpixel) >> 8) & 0xff) + chr(endpixel & 0xff)
            for color in rgb:
                r, g, b = color
                pack += chr(r) + chr(g) + chr(b)
    #            pack += chr(r & 0xff) + chr(g & 0xff) + chr(b & 0xff)
            return pack


    def _send_packet(self, p_packet):
        """Sends the given udp_packet to this's host:port"""
        try:
            self.udp_socket.sendto(p_packet, (self.udp_host, self.udp_port))
        except socket.error:
            pass

    def _reset_buffer(self):
        self.buffer = [(0, 0, 0) for _ in range(self.length)]

    def _send_rgb_frame(self, p_rgb_list):
        """
        Assuming the type/length have already been checked.
        """
        packet = self._generate_packet(0, len(p_rgb_list)-1, p_rgb_list)
        self._send_packet(packet)

    ####################
    ###Public methods###
    ####################

    def is_reachable(self):
        rgb_frame = [(0, 0, 0) for _ in range(self.length)]
        package = self._generate_packet(0, self.length -1, rgb_frame)
        try:
            self.udp_socket.sendto(package, (self.udp_host, self.udp_port))
            return True
        except socket.error:
            return False

    def calc_fps(self, p_fps):
        return STANDARD_WIDTH * p_fps / self.physikal_width

    def send_frame(self, p_frame):
        """
        Sends an frame to the blinkenfoo. A frame can be a hex-string,
        an hex list or an rgb list. Other types may cause erros later in the
        programm. The length of the frame must match the pixel count of this
        blinkenfoo. Example input, length = 3:
        Hex-String: "ff000000ff000000ff"
        Hex-List: ["ff0000", "00ff00", "0000ff"]
        RGB-List: [(255, 0, 0), (0, 255, 0), (0, 0, 255)]
        """
        rgb_frame = []
        if type(p_frame) is str:
            rgb_frame = self._hex_string_to_rgb_list(p_frame)
        elif type(p_frame) is list:
                if type(p_frame[0]) is tuple:
                    rgb_frame = p_frame
                else:
                    rgb_frame = self._hex_list_to_rgb_list(p_frame)
        else:
            raise ValueError("Frame is neither str or list. Given type: %s" % str(type(p_frame)))

        length = len(rgb_frame)
        if length == self.length:
            packet = self._generate_packet(0, len(rgb_frame)-1, rgb_frame)
        else:
            raise ValueError("Given list length does not match pixel count!: Given %d, expected %d" % (length, self.length))
        self._send_packet(packet)

    def set_full_color(self, p_color):
        """Displays the given color on every pixel of the blinkenfoo"""
        color = self._unify_color_to_rgb(p_color)
        rgb_list = [color for _ in range(self.length)]
        self._send_rgb_frame(rgb_list)

    def set_led(self, p_index, p_color):
        """Sets an specific color at the given index. The change is visible after
        flush is called"""
        if p_index > self.length:
            raise ValueError("Out of range: given %d, length %d" % (p_index, self.length))
        p_color = self._unify_color_to_rgb(p_color)
        self.buffer[p_index] = p_color

    def flush(self, p_reset = True):
        self._send_rgb_frame(self.buffer)
        if p_reset:
            self._reset_buffer()

##################
###End of class###
##################


spheres = Blinkenfoo("Spheres", "ESP_35D447.warpzone", 9, 20)
panel = Blinkenfoo("Panel", "ESP_35d9E4.warpzone", 8, 12)
warp_sign = Blinkenfoo("Warp-Sign", "ESP_133C4C.warpzone", 1, 25)
dmx = Blinkenfoo("DMX", "10.0.3.27", 5, 10)
cubes = Blinkenfoo("Cubes", "cubes.warpzone", 8, 5)
